/*
 * $Id: xm2mod.c,v 1.1 2000/11/05 17:43:41 claudio2 Exp $
 *
 * A quick-and-dirty utility to convert [unix] soundtracker xm files to
 * standard [amiga] protracker mods used in xrally.
 *
 * Parts recycled from xmp.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define PACKED __attribute__((packed))
typedef char int8;
typedef short int16;
typedef int int32;
typedef unsigned char uint8;
typedef unsigned short uint16;
typedef unsigned int uint32;

#include "xm.h"
#include "mod.h"

/* Endianism fixup */
#define FIX_ENDIANISM_16(x)	(x=((((x)&0xff00)>>8)|(((x)&0xff)<<8)))
#define FIX_ENDIANISM_32(x)	(x=(((x)&0xff000000)>>24)|(((x)&0xff0000)>>8)|\
				   (((x)&0xff00)<<8)|(((x)&0xff)<<24))
#ifdef WORDS_BIGENDIAN
#define L_ENDIAN16(x)	FIX_ENDIANISM_16(x)
#define L_ENDIAN32(x)	FIX_ENDIANISM_32(x)
#define B_ENDIAN16(x)	(x=x)
#define B_ENDIAN32(x)	(x=x)
#else
#define L_ENDIAN16(x)	(x=x)
#define L_ENDIAN32(x)	(x=x)
#define B_ENDIAN16(x)	FIX_ENDIANISM_16(x)
#define B_ENDIAN32(x)	FIX_ENDIANISM_32(x)
#endif

/* Amiga periods */
static int period_amiga[] = {
   /*  0       1       2       3       4       5       6       7   */
    0x1c56, 0x1c22, 0x1bee, 0x1bbb, 0x1b87, 0x1b55, 0x1b22, 0x1af0,  /* B  */
    0x1abf, 0x1a8e, 0x1a5d, 0x1a2c, 0x19fc, 0x19cc, 0x199c, 0x196d,  /* C  */
    0x193e, 0x1910, 0x18e2, 0x18b4, 0x1886, 0x1859, 0x182c, 0x1800,  /* C# */
    0x17d4, 0x17a8, 0x177c, 0x1751, 0x1726, 0x16fb, 0x16d1, 0x16a7,  /* D  */
    0x167d, 0x1654, 0x162b, 0x1602, 0x15d9, 0x15b1, 0x1589, 0x1562,  /* D# */
    0x153a, 0x1513, 0x14ec, 0x14c6, 0x149f, 0x1479, 0x1454, 0x142e,  /* E  */
    0x1409, 0x13e4, 0x13c0, 0x139b, 0x1377, 0x1353, 0x1330, 0x130c,  /* F  */
    0x12e9, 0x12c6, 0x12a4, 0x1282, 0x125f, 0x123e, 0x121c, 0x11fb,  /* F# */
    0x11da, 0x11b9, 0x1198, 0x1178, 0x1157, 0x1137, 0x1118, 0x10f8,  /* G  */
    0x10d9, 0x10ba, 0x109b, 0x107d, 0x105e, 0x1040, 0x1022, 0x1004,  /* G# */
    0x0fe7, 0x0fca, 0x0fad, 0x0f90, 0x0f73, 0x0f57, 0x0f3a, 0x0f1e,  /* A  */
    0x0f02, 0x0ee7, 0x0ecb, 0x0eb0, 0x0e95, 0x0e7a, 0x0e5f, 0x0e45,  /* A# */
    0x0e2b, 0x0e11, 0x0df7, 0x0ddd, 0x0dc3, 0x0daa, 0x0d91, 0x0d78,  /* B  */
};


/* Get period from note and finetune. Don't ask. */
static int note_to_period (int n, int b)
{
    int *t = period_amiga;
    int f = ((b % 100) << 7) / 100;

    if ((n += !((b < 0) && (f += 0x80)) + b / 100) < 0)
	n = 0;
    t += ((n % 12) << 3) + (f >> 4);
    return ((*t << 4) + (*t++ - *t) * (f & 0xf)) >> (n / 12);
}


/* Convert differential to absolute sample data */
static void diff2abs (int l, int r, char *p)
{
	uint16* w = (uint16*) p;
	uint16 abs = 0;

	if (r) {
		for (l >>= 1; l--; ) {
			abs = *w + abs;
			*w++ = abs;
		}
	} else {
		for (; l--; ) {
			abs = *p + abs;
			*p++ = (uint8) abs;
		}
	}
}

/* Core business */
static int xm2mod (FILE *in, FILE *out)
{
	struct mod_header mh;
	struct xm_file_header fh;
	struct xm_header xh;
	int i, j, k;

	fread (&fh, 1, sizeof (struct xm_file_header), in);
	if (memcmp (fh.id, "Extended Module: ", 17) != 0)
		return -1;

	fread (&xh, 1, sizeof (struct xm_header), in);

	printf ("Converting \"%s\"\n", fh.name);
	memset (&mh, 0, sizeof (struct mod_header));
	memcpy (mh.name, fh.name, 20);
	mh.len = xh.songlen;
	memcpy (mh.order, xh.order, 128);
	memcpy (mh.magic, "M.K.", 4);
	
	printf ("Writing header\n");
	fwrite (&mh, 1, sizeof (struct mod_header), out);

	printf ("Writing patterns...");
	for (i = 0; i < xh.patterns; i++) {
		struct xm_pattern_header xph;
		uint8 event[4];
		int n;

		fread (&xph, 1, sizeof (struct xm_pattern_header), in);

		for (j = 0; j < 64; j++) {
			for (k = 0; k < xh.channels; k++) {
				int note, ins, vol, fxp, fxt;
	
				/* Fetch event */
				note = fgetc (in);
				vol = ins = fxt = fxp = 0;
				if (note & XM_EVENT_PACKING) {
					int pack = note & XM_EVENT_PACK_MASK;
					note = 0;
					if (pack & XM_EVENT_NOTE_FOLLOWS)
						note = fgetc (in);
					if (pack & XM_EVENT_INSTRUMENT_FOLLOWS)
						ins = fgetc (in);
					if (pack & XM_EVENT_VOLUME_FOLLOWS)
						vol = fgetc (in);
					if (pack & XM_EVENT_FXTYPE_FOLLOWS)
						fxt = fgetc (in);
					if (pack & XM_EVENT_FXPARM_FOLLOWS)
						fxp = fgetc (in);
				} else {
					ins = fgetc (in);
					vol = fgetc (in);
					fxt = fgetc (in);
					fxp = fgetc (in);
				}

				n = note ? note_to_period (note - 2, 0) >> 4: 0;
				event[0] = (ins & 0xf0) | ((n & 0xf00) >> 8);
				event[1] = n & 0xff;
				event[2] = ((ins & 0x0f) << 4) | (fxt & 0x0f);
				event[3] = fxp;

				fwrite (&event, 1, 4, out);
			}
		}
		printf (" %d", i);
	}

	printf ("\nWriting samples...");
	for (i = 0; i < xh.instruments; i++) {
		struct xm_instrument_header xih;
		struct xm_instrument xi;
		struct xm_sample_header xs;
		int sdata;

		fread (&xih, 1, sizeof (struct xm_instrument_header), in);
		if (xih.samples == 0) {
			mh.ins[i].size = 0;
			mh.ins[i].loop_start = 0;
			mh.ins[i].loop_size = 1;
			continue;
		}
		fread (&xi, 1, sizeof (struct xm_instrument), in);

		sdata = 0;
		for (j = 0; j < xih.samples; j++) {
			fread (&xs, 1, sizeof (struct xm_sample_header), in);
			if (j == 0) {
				int8 *buf;
				mh.ins[i].size = xs.length >> 1;
				mh.ins[i].loop_start = xs.loop_start >> 1;
				mh.ins[i].loop_size = xs.loop_length >> 1;
				mh.ins[i].volume = xs.volume;
				B_ENDIAN16 (mh.ins[i].size);
				B_ENDIAN16 (mh.ins[i].loop_start);
				B_ENDIAN16 (mh.ins[i].loop_size);
				buf = malloc (xs.length);
				fread (buf, 1, xs.length, in);
				diff2abs (xs.length, 0, buf);
				fwrite (buf, 1, xs.length, out);
			} else {
				sdata += xs.length *
					((xs.type & XM_SAMPLE_16BIT) ? 2 : 1);
			}
		}
		fseek (in, sdata, SEEK_CUR);
		printf (" %d", i);
	}

	fseek (out, 0, SEEK_SET);
	fwrite (&mh, 1, sizeof (struct mod_header), out);
	printf ("\nDone.\n");

	return 0;
}


int main (int argc, char **argv)
{
	FILE *in, *out;

	if (argc != 3) {
		fprintf (stderr, "Usage: xm2mod foo.xm bar.mod\n");
		exit (-1);
	}
	
	in = fopen (argv[1], "r");
	out = fopen (argv[2], "w");

	if (in == NULL || out == NULL) {
		fprintf (stderr, "Huh?");
		exit (-1);
	}

	setbuf (stdout, NULL);
	xm2mod (in, out);

	return 0;
}

